"""Base support code for Kubernetes template generation."""
from typing import Any, List

import yaml

from templatizer import NoValue, Templatable


def fix_value(val: Any):
    """fix_value fixes value types to embed into YAML correctly."""
    if isinstance(val, list):
        return [fix_value(v) for v in val]

    # Other templatable objects should be emitted correctly
    if isinstance(val, Templatable):
        val = yaml.load(val.generate(), Loader=yaml.Loader)

    return val


class K8STemplatable(Templatable):
    """K8STemplatable is the base class for all Kubernetes template objects."""

    description = NoValue
    apiVersion = NoValue
    kind = NoValue
    required_props: List[str] = []

    def generate(self) -> str:
        document = {}

        props = self.propval("props")
        required_props = self.propval("required_props")
        for prop in props:
            val = self.propval(prop)
            # K8S-specific feature: returning None also ignores the value
            if val is not NoValue and val:
                val = fix_value(val)
                document[prop] = val
            elif prop in required_props:
                raise ValueError(f'no value for required property "%{prop}s"')

        return yaml.dump(document)



"""Functions for running the generator."""
from typing import Iterable

from .templatable import Templatable


def run(objects: Iterable[Templatable], separator: str = "---"):
    """Run the generator on the given objects, adding a separator for each."""
    return separator.join(obj.generate() for obj in objects)



"""Entry point for the templatizer package."""
from .run import run
from .templatable import NoValue, Templatable
from .yamlblob import YamlBlob

__all__ = ["run", "Templatable", "YamlBlob", "NoValue"]



"""Provides YAML 'blob' template classes for embedding YAML as strings."""
from typing import Any

import yaml

from .templatable import Templatable


def string_as_block(dumper, data):
    """string_as_block uses YAML's block style for strings with newlines."""
    if "\n" in data:
        return dumper.represent_scalar("tag:yaml.org,2002:str", data, style="|")
    return dumper.represent_scalar("tag:yaml.org,2002:str", data)


yaml.add_representer(str, string_as_block)


class YamlBlob(Templatable):
    """YamlBlob embeds data as a YAML string."""

    def data(self) -> Any:
        """data returns the raw data to be YAML-encoded."""
        raise NotImplementedError("data() must be implemented on ConfigurationData")

    def generate(self) -> str:
        return yaml.dump(self.data())



"""The base class and utilities for templateable objects."""
from typing import Any

# Global NoValue can be used to omit a value
# (because a function that does not return any value returns None)
# To check, you can use `foo is NoValue`
NoValue = object()


class Templatable:
    """An object that is able to be converted into a template."""

    def generate(self) -> str:
        """generate generates a template from the Templatable."""
        raise NotImplementedError("generate() must be implemented")

    def propval(self, k: str) -> Any:
        """propval retrieves the property value for the given key."""
        try:
            val = getattr(self, k)
        except AttributeError:
            val = NoValue

        if val is NoValue:
            return val

        if callable(val):
            return val()

        return val
